/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Copyright 1997 - July 2008 CWI, August 2008 - 2018 MonetDB B.V.
 */

#ifndef STORE_DEPEND_H
#define STORE_DEPEND_H

#include "sql_types.h"
#include "sql_backend.h"
#include "store_sequence.h"

#endif /*STORE_DEPEND_H */
